@@ -5,26 +5,38 @@
 # SPDX-License-Identifier: AFL-3.0
 #
 
-#!/usr/bin/python3.6
+'''Module for managing the file based database backend (with TinyDB)'''
 
-from snooze.db.database import Database
-from snooze.utils.functions import dig, flatten, to_tuple
-from threading import Lock
-from logging import getLogger
-from datetime import datetime
-from bson.json_util import dumps
-from pathlib import Path
-import uuid
 import re
-log = getLogger('snooze.db.file')
-
-from tinydb import TinyDB, Query as BaseQuery
 from copy import deepcopy
+from datetime import datetime
 from functools import reduce
+from logging import getLogger
+from pathlib import Path
+from threading import Lock
+from typing import List, Tuple
 
-class OperationNotSupported(Exception): pass
+import uuid
+from bson.json_util import dumps
+from tinydb import TinyDB, Query as BaseQuery, where
+from tinydb.operations import add
+
+from snooze.db.database import Database, wrap_exception
+from snooze.utils.config import FileConfig
+from snooze.utils.functions import dig, to_tuple, flatten
+
+log = getLogger('snooze.db.file')
+DEFAULT_PAGINATION = {
+    'orderby': '',
+    'page_number': 1,
+    'nb_per_page': 0,
+    'asc': True,
+    'only_one': False,
+}
+
+class OperationNotSupported(Exception):
+    '''Raised when the search operator is not supported'''
 
-default_filename = 'db.json'
 mutex = Lock()
 
 class Query(BaseQuery):
@@ -54,40 +66,84 @@ def test_contains(array, value):
 def test_search(dic, value):
     return value in str(dic)
 
+def append_many(fields):
+    def transform(doc):
+        for field, val in fields.items():
+            doc[field] += val
+    return transform
+
+def prepend_many(fields):
+    def transform(doc):
+        for field, val in fields.items():
+            doc[field] = val + doc[field]
+    return transform
+
+def set_many(fields):
+    def transform(doc):
+        for field, val in fields.items():
+            doc[field] = val
+    return transform
+
+def remove_many(fields):
+    def transform(doc):
+        for field, val in fields.items():
+            doc[field] = [i for i in doc[field] if i not in val]
+    return transform
+
+def overwrite(data, field):
+    def transform(doc):
+        if doc[field] in data:
+            for k,v in data[doc[field]].items():
+                doc[k] = v
+    return transform
+
 class BackendDB(Database):
-    def init_db(self, conf):
-        if conf.get('path'):
-            filename = conf.get('path')
-        else:
-            filename = default_filename
-        self.db = TinyDB(filename)
-        log.debug("Initialized TinyDB at path {}".format(filename))
-        log.debug("db: {}".format(self.db))
-        log.debug("List of collections: {}".format(self.db.tables()))
+    '''Backend database based on local file (TinyDB)'''
+
+    name = 'file'
+
+    def __init__(self, config: FileConfig):
+        self.db = TinyDB(config.path)
+        log.debug("Initialized TinyDB at path %s", config.path)
+        log.debug("db: %s", self.db)
+        log.debug("List of collections: %s", self.db.tables())
 
     def create_index(self, collection, fields):
         pass
 
     def cleanup_timeout(self, collection):
         mutex.acquire()
-        #log.debug("Cleanup collection {}".format(collection))
         now = datetime.now().timestamp()
         aggregate_results = self.db.table(collection).search(Query().ttl >= 0)
-        aggregate_results = list(map(lambda doc: {'_id': doc.doc_id, 'timeout': doc['ttl'] + doc['date_epoch']}, aggregate_results))
-        aggregate_results = list(filter(lambda doc: doc['timeout'] <= now, aggregate_results))
+        aggregate_results = [{'_id': doc.doc_id, 'timeout': doc['ttl'] + doc['date_epoch']} for doc in aggregate_results]
+        aggregate_results = [doc for doc in aggregate_results if doc['timeout'] <= now]
         res = self.delete_aggregates(collection, aggregate_results)
         mutex.release()
         return res
 
-    def cleanup_orphans(self, collection, key, col_ref, key_ref):
-        mutex.acquire()
-        #log.debug("Cleanup collection {} by finding {} in collection {} matching {}".format(collection, key, col_ref, key_ref))
-        results = list(map(lambda doc: doc[key_ref], self.db.table(col_ref).all()))
-        aggregate_results = self.db.table(collection).search(~ (Query()[key].one_of(results)))
-        aggregate_results = list(map(lambda doc: {'_id': doc.doc_id}, aggregate_results))
-        res = self.delete_aggregates(collection, aggregate_results)
-        mutex.release()
-        return res
+    def cleanup_comments(self):
+        '''Delete comments which record doesn't exist anymore'''
+        with mutex:
+            record_uids = [doc['uid'] for doc in self.db.table('record').all()]
+            aggregate_results = self.db.table('comment').search(~ (Query()['record_uid'].one_of(record_uids)))
+            aggregate_results = [{'_id': doc.doc_id} for doc in aggregate_results]
+            res = self.delete_aggregates('comment', aggregate_results)
+            return res
+
+    def cleanup_orphans(self, collection: str) -> int:
+        '''Delete objects which one of their ancestors does not exist anymore'''
+        with mutex:
+            parents = set(flatten([doc.get('parents', []) for doc in self.db.table(collection).all()]))
+        if len(parents) == 0:
+           return 0
+        to_delete = []
+        for parent in parents:
+            if not self.get_one(collection, {'uid': parent}):
+                to_delete.append(parent)
+        total = 0
+        with mutex:
+            total = len(self.db.table(collection).remove(Query().parents.any(to_delete)))
+        return total
 
     def cleanup_audit_logs(self, interval):
         mutex.acquire()
@@ -112,24 +168,36 @@ class BackendDB(Database):
         log.debug("Found audit log to remove for %d objects", len(oids))
         doc_ids = [
             obj.doc_id
-            for obj in self.db.table('audit').search(Query().object_id == oid)
             for oid in oids
+            for obj in self.db.table('audit').search(Query().object_id == oid)
         ]
         log.debug("Found %d audit logs to remove", len(doc_ids))
         self.db.table('audit').remove(doc_ids=doc_ids)
 
         mutex.release()
 
+    def renumber_field(self, collection, field):
+        '''Renumber field by ascending order'''
+        log.info("Reordering field '%s' in collection %s", field, collection)
+        data = self.search(collection, pagination={'orderby': field, 'asc': True})['data']
+        rows_by_uid = {}
+        for idx, row in enumerate(data):
+            row[field] = idx
+            rows_by_uid[row['uid']] = row
+        self.db.table(collection).update(overwrite(rows_by_uid, 'uid'))
+        log.info("Field '%s' renumbering on collection %s: Success", field, collection)
+
     def delete_aggregates(self, collection, aggregate_results):
-        ids = list(map(lambda doc: doc['_id'], aggregate_results))
+        ids = [doc['_id'] for doc in aggregate_results]
         deleted_count = 0
         if ids:
             deleted_results = self.db.table(collection).remove(doc_ids=ids)
             deleted_count = len(deleted_results)
-            log.debug('Removed {} documents in {}'.format(deleted_count, collection))
+            log.debug('Removed %d documents in %s', deleted_count, collection)
         return deleted_count
 
-    def write(self, collection, obj, primary = None, duplicate_policy='update', update_time=True, constant=None):
+    @wrap_exception
+    def write(self, collection, obj, primary=None, duplicate_policy='update', update_time=True, constant=None):
         mutex.acquire()
         added = []
         updated = []
@@ -140,7 +208,7 @@ class BackendDB(Database):
         add_obj = False
         table = self.db.table(collection)
         tobj = deepcopy(obj)
-        if type(tobj) != list:
+        if not isinstance(tobj, list):
             tobj = [tobj]
         if primary:
             if isinstance(primary , str):
@@ -159,7 +227,7 @@ class BackendDB(Database):
                 primary_query = reduce(lambda a, b: a & b, primary_query)
                 primary_docs = table.search(primary_query)
                 if primary_docs:
-                    log.debug('Documents with same primary {}: {}'.format(primary, str(primary_docs[0].doc_id)))
+                    log.debug('Documents with same primary %s: %s', primary, primary_docs[0].doc_id)
             if 'uid' in o:
                 query = Query()
                 docs = table.search(query.uid == o['uid'])
@@ -167,24 +235,26 @@ class BackendDB(Database):
                     doc = docs[0]
                     doc_id = doc.doc_id
                     old = doc
-                    log.debug('Found: {}'.format(str(doc_id)))
+                    log.debug('Found: %s', doc_id)
                     if primary_docs and doc_id != primary_docs[0].doc_id:
-                        error_message = f"Found another document with same primary {primary}: {primary_docs}. Since UID is different, cannot update"
+                        error_message = f"Found another document with same primary {primary}: {primary_docs}. " \
+                            "Since UID is different, cannot update"
                         log.error(error_message)
                         o['error'] = error_message
                         rejected.append(o)
                     elif constant and any(doc.get(c, '') != o.get(c) for c in constant):
-                        error_message = f"Found a document with existing uid {o['uid']} but different constant values: {constant}. Since UID is different, cannot update"
+                        error_message = f"Found a document with existing uid {o['uid']} but different constant " \
+                            f"values: {constant}. Since UID is different, cannot update"
                         log.error(error_message)
                         o['error'] = error_message
                         rejected.append(o)
                     elif duplicate_policy == 'replace':
-                        log.debug('Replacing with: {}'.format(str(doc_id)))
+                        log.debug('Replacing with: %s', doc_id)
                         table.remove(doc_ids=[doc_id])
                         table.insert(o)
                         replaced.append(o)
                     else:
-                        log.debug('Updating with: {}'.format(str(doc_id)))
+                        log.debug('Updating with: %s', doc_id)
                         table.update(o, doc_ids=[doc_id])
                         updated.append(o)
                 else:
@@ -198,12 +268,13 @@ class BackendDB(Database):
                     doc_id = doc.doc_id
                     old = doc
                     if constant and any(doc.get(c, '') != o.get(c) for c in constant):
-                        error_message = f"Found a document with existing primary {primary} but different constant values: {constant}. Since UID is different, cannot update"
+                        error_message = f"Found a document with existing primary {primary} but different "\
+                            f"constant values: {constant}. Since UID is different, cannot update"
                         log.error(error_message)
                         o['error'] = error_message
                         rejected.append(o)
                     else:
-                        log.debug('Evaluating duplicate policy: {}'.format(duplicate_policy))
+                        log.debug('Evaluating duplicate policy: %s', duplicate_policy)
                         if duplicate_policy == 'insert':
                             add_obj = True
                         elif duplicate_policy == 'reject':
@@ -211,14 +282,14 @@ class BackendDB(Database):
                             o['error'] = error_message
                             rejected.append(o)
                         elif duplicate_policy == 'replace':
-                            log.debug('Replace with: {}'.format(str(doc_id)))
+                            log.debug('Replace with: %s', doc_id)
                             table.remove(doc_ids=[doc_id])
                             if 'uid' in doc:
                                 o['uid'] = doc['uid']
                             table.insert(o)
                             replaced.append(o)
                         else:
-                            log.debug('Update with: {}'.format(str(doc_id)))
+                            log.debug('Update with: %s', doc_id)
                             table.update(o, doc_ids=[doc_id])
                             updated.append(o)
                 else:
@@ -231,14 +302,82 @@ class BackendDB(Database):
                 obj_copy[-1]['uid'] = str(uuid.uuid4())
                 added.append(o)
                 add_obj = False
-                log.debug("In {}, inserting {}".format(collection, o.get('uid', '')))
+                log.debug("In %s, inserting %s", collection, o.get('uid', ''))
             if old:
                 o['_old'] = old
         if len(obj_copy) > 0:
             table.insert_multiple(obj_copy)
         mutex.release()
-        return {'data': {'added': deepcopy(added), 'updated': deepcopy(updated), 'replaced': deepcopy(replaced),'rejected': deepcopy(rejected)}}
+        return {
+            'data': {
+                'added': deepcopy(added),
+                'updated': deepcopy(updated),
+                'replaced': deepcopy(replaced),
+                'rejected': deepcopy(rejected),
+            },
+        }
 
+    @wrap_exception
+    def get_one(self, collection: str, search: dict):
+        '''Return the first element found based on a search dict'''
+        with mutex:
+            queries = []
+            for key, value in search.items():
+                query = dig(Query(), *key.split('.')) == value
+                queries.append(query)
+            search_query = reduce(lambda a, b: a & b, queries)
+            results = self.db.table(collection).search(search_query)
+            if results:
+                return results[0]
+            else:
+                return None
+
+    @wrap_exception
+    def replace_one(self, collection: str, search: dict, obj: dict, update_time: bool = True):
+        with mutex:
+            query = Query()
+            new_obj = dict(obj)
+            queries = []
+            for key, value in search.items():
+                new_obj[key] = value
+                query = dig(Query(), *key.split('.')) == value
+                queries.append(query)
+            if update_time:
+                new_obj['date_epoch'] = datetime.now().timestamp()
+            search_query = reduce(lambda a, b: a & b, queries)
+            search = self.db.table(collection).search(search_query)
+            if search:
+                self.db.table(collection).remove(doc_ids=[search[0].doc_id])
+            self.db.table(collection).insert(new_obj)
+
+    @wrap_exception
+    def update_one(self, collection: str, uid: str, obj: dict, update_time: bool = True):
+        with mutex:
+            new_obj = dict(obj)
+            if update_time:
+                new_obj['date_epoch'] = datetime.now().timestamp()
+            self.db.table(collection).upsert(new_obj, where('uid') == uid)
+
+    @wrap_exception
+    def bulk_increment(self, collection: str, updates: List[Tuple[dict, dict]], upsert: bool = False):
+        '''Perform a bulk update of increments. Each update should be a tuple of search and update'''
+        with mutex:
+            for search, update in updates:
+                queries = []
+                for key, value in search.items():
+                    if isinstance(value, datetime):
+                        value = int((value.timestamp() // 3600) * 3600)
+                        search[key] = value
+                    query = dig(Query(), *key.split('.')) == value
+                    queries.append(query)
+                search_query = reduce(lambda a, b: a & b, queries)
+                update.pop('_id', None)
+                for key, value in update.items():
+                    self.db.table(collection).update(add(key, value), search_query)
+                if upsert and not self.db.table(collection).search(search_query):
+                    self.db.table(collection).insert({**search, **update})
+
+    @wrap_exception
     def inc(self, collection, field, labels={}):
         now = int((datetime.now().timestamp() // 3600) * 3600)
         mutex.acquire()
@@ -248,8 +387,8 @@ class BackendDB(Database):
         added = []
         updated = []
         if labels:
-            for k,v in labels.items():
-                keys.append(field+'__'+k+'__'+v)
+            for key, value in labels.items():
+                keys.append(f"{field}__{key}__{value}")
         else:
             keys.append(field)
         for key in keys:
@@ -267,30 +406,52 @@ class BackendDB(Database):
         mutex.release()
         return {'data': {'added': added, 'updated': updated}}
 
-    def update_fields(self, collection, fields, condition=[]):
-        log.debug("Update collection '{}' with fields '{}' based on the following search".format(collection, fields))
+    def inc_many(self, collection: str, field: str, condition = None, value: int = 1):
+        if condition is None:
+            condition = []
+        tinydb_search = self.convert(condition)
         total = 0
-        mutex.acquire()
-        if collection not in self.db.tables():
-            mutex.release()
-            return 0
-        mutex.release()
-        results = self.search(collection, condition)
-        total = results['count']
-        for record in results['data']:
-            for field, val in fields.items():
-                record[field] = val
-        if total > 0:
-            self.write(collection, results['data'])
-        log.debug("Updated {} fields".format(total))
+        if collection in self.db.tables():
+            table = self.db.table(collection)
+            total = table.count(tinydb_search)
+            table.update(add(field, value), tinydb_search)
         return total
 
+    def update_with_operation(self, collection, operation, condition=[]):
+        tinydb_search = self.convert(condition)
+        total = 0
+        mutex.acquire()
+        if collection in self.db.tables():
+            try:
+                total = len(self.db.table(collection).update(operation, tinydb_search))
+            except Exception as err:
+                log.exception(err)
+        mutex.release()
+        log.debug("Updated %d document(s)", total)
+        return total
+
+    def set_fields(self, collection, fields, condition=[]):
+        log.debug("Update collection '%s' with fields '%s' based on the following search '%s'", collection, fields, condition)
+        return self.update_with_operation(collection, set_many(fields), condition)
+
+    def append_list(self, collection, fields, condition=[]):
+        log.debug("Append to collection '%s' fields '%s' based on the following search: '%s'", collection, fields, condition)
+        return self.update_with_operation(collection, append_many(fields), condition)
+
+    def prepend_list(self, collection, fields, condition=[]):
+        log.debug("Prepend to collection '%s' fields '%s' based on the following search: '%s'", collection, fields, condition)
+        return self.update_with_operation(collection, prepend_many(fields), condition)
+
+    def remove_list(self, collection, fields, condition=[]):
+        log.debug("Remove from collection '%s' fields '%s' based on the following search: '%s'", collection, fields, condition)
+        return self.update_with_operation(collection, remove_many(fields), condition)
+
     def compute_stats(self, collection, date_from, date_until, groupby='hour'):
-        log.debug("Compute metrics on `{}` from {} until {} grouped by {}".format(collection, date_from, date_until, groupby))
+        log.debug("Compute metrics on `%s` from %s until %s grouped by %s", collection, date_from, date_until, groupby)
         date_from = date_from.replace(minute=0, second=0, microsecond=0)
         mutex.acquire()
         if collection not in self.db.tables():
-            log.debug("Compute stats: collection {} does not exist".format(collection))
+            log.debug("Compute stats: collection %s does not exist", collection)
             mutex.release()
             return {'data': [], 'count': 0}
         if groupby == 'hour':
@@ -324,51 +485,62 @@ class BackendDB(Database):
             elif doc['key'] not in groups[date_range]:
                 groups[date_range][doc['key']] = {'value': 0}
             groups[date_range][doc['key']]['value'] += doc['value']
-        for date, v in groups.items():
+        for date, value in groups.items():
             entry = {'_id': date, 'data': []}
-            for key, doc in v.items():
+            for key, doc in value.items():
                 entry['data'].append({'key': key, 'value': doc['value']})
             res.append(entry)
         results_agg = sorted(res, key=lambda d: d['_id'])
         count = len(results_agg)
-        log.debug("Compute stats: Got {} results".format(count))
+        log.debug("Compute stats: Got %s results", count)
         mutex.release()
         return {'data': results_agg, 'count': count}
 
-    def search(self, collection, condition=[], nb_per_page=0, page_number=1, orderby="", asc=True):
+    @wrap_exception
+    def search(self, collection, condition=[], **pagination):
         mutex.acquire()
+        pagination = {**DEFAULT_PAGINATION, **pagination}
+        orderby = pagination['orderby']
+        page_number = pagination['page_number']
+        nb_per_page = pagination['nb_per_page']
+        asc = pagination['asc']
+        only_one = pagination['only_one']
         tinydb_search = self.convert(condition)
-        #log.debug("Condition {} converted to tinydb search {}".format(condition, tinydb_search))
         if collection in self.db.tables():
             table = self.db.table(collection)
             if tinydb_search:
                 results = table.search(tinydb_search)
             else:
                 results = table.all()
-            total = len(results)
-            if nb_per_page > 0:
-                from_el = max((page_number-1)*nb_per_page, 0)
-                to_el = page_number*nb_per_page
-            else:
-                from_el = None
-                to_el = None
             if len(orderby) > 0 and all(dig(res, *orderby.split('.')) for res in list(results)):
                 results = sorted(list(results),  key=lambda x: reduce(lambda c, k: c.get(k, {}), orderby.split('.'), x))
             if not asc:
                 results = list(reversed(results))
-            results = results[from_el:to_el]
-            log.debug("Found {} result(s) for search {} in collection {}. Page: {}-{}. Sort by {}. Order: {}".format(total, tinydb_search, collection, page_number, nb_per_page, orderby, 'Ascending' if asc else 'Descending'))
+            if only_one:
+                total = 1
+                results = results[:1]
+            else:
+                total = len(results)
+                if nb_per_page > 0:
+                    from_el = max((page_number-1)*nb_per_page, 0)
+                    to_el = page_number*nb_per_page
+                else:
+                    from_el = None
+                    to_el = None
+                results = results[from_el:to_el]
+            log.debug("Found %d result(s) for search %s in collection %s. Pagination: %s",
+                total, tinydb_search, collection, pagination)
             mutex.release()
             return {'data': deepcopy(results), 'count': total}
         else:
-            log.warning("Cannot find collection {}".format(collection))
+            log.warning("Cannot find collection %s", collection)
             mutex.release()
             return {'data': [], 'count': 0}
 
+    @wrap_exception
     def delete(self, collection, condition=[], force=False):
         mutex.acquire()
         tinydb_search = self.convert(condition)
-        #log.debug("Condition {} converted to tinydb delete search {}".format(condition, tinydb_search))
         if collection in self.db.tables():
             table = self.db.table(collection)
             if len(condition) == 0 and not force:
@@ -381,14 +553,19 @@ class BackendDB(Database):
                 else:
                     results = table.remove(tinydb_search)
                     results_count = len(results)
-                log.debug("Found {} item(s) to delete in collection {} for search {}".format(results_count, collection, tinydb_search))
+                log.debug("Found %d item(s) to delete in collection %s for search %s",
+                    results_count, collection, tinydb_search)
             mutex.release()
             return {'data': [], 'count': results_count}
         else:
             mutex.release()
-            log.error("Cannot find collection {}".format(collection))
+            log.error("Cannot find collection %s", collection)
             return {'data': 0}
 
+    def drop(self, collection):
+        if collection in self.db.tables():
+            self.db.drop_table(collection)
+
     def convert(self, array):
         """
         Convert `Condition` type from snooze.utils
@@ -445,7 +622,7 @@ class BackendDB(Database):
                     saved_key = key
                     key = self.convert(key)
                     converted = True
-                except:
+                except Exception:
                     key = saved_key
             if converted:
                 return_obj = dig(Query(), *value.split('.')).any(key)
@@ -457,16 +634,16 @@ class BackendDB(Database):
             arg = args[0]
             try:
                 return_obj = Query().test_root(test_search, to_tuple(arg))
-            except Exception as e:
-                log.exception(e)
+            except Exception as err:
+                log.exception(err)
                 raise OperationNotSupported(operation)
         else:
             raise OperationNotSupported(operation)
         return return_obj
 
-    def backup(self, backup_path, backup_exclude = []):
+    def backup(self, backup_path, backup_exclude=[]):
         collections = [c for c in self.db.tables() if c not in backup_exclude]
-        log.debug('Starting backup of {}'.format(collections))
+        log.debug('Starting backup of %s', collections)
         succeeded = []
         for i, collection_name in enumerate(collections):
             try:
@@ -475,10 +652,7 @@ class BackendDB(Database):
                 with jsonpath.open("wb") as jsonfile:
                     jsonfile.write(dumps(collection).encode())
                     succeeded.append(collection_name)
-            except Exception as e:
-                log.error('Backup of {} failed'.format(collection_name))
-                log.exception(e)
-        log.info('Backup of {} succeeded'.format(succeeded))
-
-    def get_uri(self):
-        return None
+            except Exception as err:
+                log.error('Backup of %s failed', collection_name)
+                log.exception(err)
+        log.info('Backup of %s succeeded', succeeded)
