@@ -1,18 +1,28 @@
-#!/usr/bin/python
-from logging import getLogger
-log = getLogger('snooze.webhooks.influxdb')
+#
+# Copyright 2018-2020 Florian Dematraz <florian.dematraz@snoozeweb.net>
+# Copyright 2018-2020 Guillaume Ludinard <guillaume.ludi@gmail.com>
+# Copyright 2020-2021 Japannext Co., Ltd. <https://www.japannext.co.jp/>
+# SPDX-License-Identifier: AFL-3.0
+#
+
+'''A webhook for influxdb v2 notifications'''
 
 from copy import deepcopy
-from snooze.api.falcon import WebhookRoute
-from snooze.utils.functions import sanitize
+from logging import getLogger
+
 from bson.json_util import loads
 
+from snooze.api.routes import WebhookRoute
+from snooze.utils.functions import sanitize
+
+log = getLogger('snooze.webhooks.influxdb')
+
 class InfluxDBRoute(WebhookRoute):
-    auth = {
-        'auth_disabled': True
-    }
+    '''A falcon route to handle InfluxDB v2 alerts'''
+    authentication = False
 
     def parse(self, media):
+        '''Parse the data of the webhook to create an alert'''
         alert = {}
 
         media = sanitize(media)
@@ -32,12 +42,12 @@ class InfluxDBRoute(WebhookRoute):
         alert['severity'] = media.get('severity', level)
         alert['message'] = media.get('_message', '')
         alert['source'] = 'influxdb2'
-        for k, v in media.items():
-            if k[0] != '_':
+        for key, value in media.items():
+            if key[0] != '_':
                 try:
-                    alert[k] = sanitize(loads(v))
-                except:
-                    alert[k] = v
+                    alert[key] = sanitize(loads(value))
+                except Exception:
+                    alert[key] = value
 
         return alert
 
