@@ -5,14 +5,16 @@
 # SPDX-License-Identifier: AFL-3.0
 #
 
-#!/usr/bin/python3.6
+'''Plugin definition for the KV'''
+
+from logging import getLogger
 
 from snooze.plugins.core import Plugin
-import logging
-from logging import getLogger
+
 log = getLogger('snooze.kv')
 
 class Kv(Plugin):
+    '''Plugin for managing a user/script defined key-value in the database'''
     def reload_data(self, sync = False):
         super().reload_data()
         kv = {}
@@ -21,12 +23,13 @@ class Kv(Plugin):
                 if key_val['dict'] not in kv:
                     kv[key_val['dict']] = {}
                 kv[key_val['dict']][key_val['key']] = key_val['value']
-            except Exception as e:
-                log.exception(e)
+            except Exception as err:
+                log.exception(err)
                 continue
         self.kv = kv
-        if sync and self.core.cluster:
-            self.core.cluster.reload_plugin(self.name)
+        if sync:
+            self.sync_neighbors()
 
     def get(self, mydict, key):
+        '''Return the value for a given dictionary and key'''
         return self.kv[mydict][key]
